/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <libxfce4util/libxfce4util.h>

#include "keybindings.h"

#define KEYBIND_FILE "xfmedia/keybindings.rc"

#define KEYBIND_HASH_KEY(keyval, modmask) \
		GUINT_TO_POINTER((keyval << 16) | (modmask & 0xffff))
#define KEYBIND_UNHASH(hashval, keyval, modmask) \
	do {\
		keyval = (hashval & 0xffff0000) >> 16;\
		modmask = hashval & 0xffff;\
	} while(0)

static GHashTable *keys_hash = NULL;

static gboolean
parse_keybinding(const gchar *str, guint *keyval, GdkModifierType *modmask)
{
	gchar **parts;
	gint i;
	gboolean ret = TRUE;
	
	g_return_val_if_fail(str && keyval && modmask, FALSE);
	
	*keyval = 0;
	*modmask = 0;
	
	parts = g_strsplit(str, "+", -1);
	for(i = 0; parts[i]; i++) {
		if(!g_ascii_strcasecmp(parts[i], "shift"))
			*modmask |= GDK_SHIFT_MASK;
		else if(!g_ascii_strcasecmp(parts[i], "control") || !g_ascii_strcasecmp(parts[i], "ctrl"))
			*modmask |= GDK_CONTROL_MASK;
		else if(!g_ascii_strcasecmp(parts[i], "alt") || !g_ascii_strcasecmp(parts[i], "mod1"))
			*modmask |= GDK_MOD1_MASK;
		else if(!g_ascii_strcasecmp(parts[i], "super") || !g_ascii_strcasecmp(parts[i], "mod4"))
			*modmask |= GDK_MOD3_MASK;
		else {
			if(strlen(parts[i]) == 1) {
				if(*parts[i] < 33 || *parts[i] > 126) {
					ret = FALSE;
					break;
				}
				*parts[i] = g_ascii_tolower(*parts[i]);
				*keyval = gdk_keyval_from_name(parts[i]);
			} else {
				if(!g_ascii_strcasecmp(parts[i], "up"))
					*keyval = GDK_Up;
				else if(!g_ascii_strcasecmp(parts[i], "down"))
					*keyval = GDK_Down;
				else if(!g_ascii_strcasecmp(parts[i], "left"))
					*keyval = GDK_Left;
				else if(!g_ascii_strcasecmp(parts[i], "right"))
					*keyval = GDK_Right;
				else if(!g_ascii_strcasecmp(parts[i], "space"))
					*keyval = GDK_space;
				else if(!g_ascii_strcasecmp(parts[i], "insert"))
					*keyval = GDK_Insert;
				else if(!g_ascii_strcasecmp(parts[i], "delete"))
					*keyval = GDK_Delete;
				else if(!g_ascii_strcasecmp(parts[i], "backspace"))
					*keyval = GDK_BackSpace;
				else if(!g_ascii_strcasecmp(parts[i], "home"))
					*keyval = GDK_Home;
				else if(!g_ascii_strcasecmp(parts[i], "end"))
					*keyval = GDK_End;
				else if(!g_ascii_strcasecmp(parts[i], "esc") || !g_ascii_strcasecmp(parts[i], "escape"))
					*keyval = GDK_Escape;
				else if(!g_ascii_strcasecmp(parts[i], "pgup") || !g_ascii_strcasecmp(parts[i], "pageup"))
					*keyval = GDK_Page_Up;
				else if(!g_ascii_strcasecmp(parts[i], "pgdown") || !g_ascii_strcasecmp(parts[i], "pagedown"))
					*keyval = GDK_Page_Down;
				else if(*parts[i] == 'f' || *parts[i] == 'F') {
					const guint funckeys[12] = {
						GDK_F1, GDK_F2, GDK_F3, GDK_F4, GDK_F5, GDK_F6,
						GDK_F7, GDK_F8, GDK_F9, GDK_F10, GDK_F11, GDK_F12
					};
					gint key = atoi(parts[i]+1);
					if(key > 0 && key < 13)
						*keyval = funckeys[key-1];
					else {
						ret = FALSE;
						break;
					}
				} else {
					ret = FALSE;
					break;
				}
			}
		}
	}
	
	if(parts)
		g_strfreev(parts);
	
	return ret;
}

static gboolean
xfmedia_keybindings_parse_file(const gchar *file)
{
	gboolean ret = FALSE;
	FILE *fp;
	gchar buf[2048], *p;
	gint line = 0;
	
	fp = fopen(file, "r");
	if(fp) {
		*buf = 0;
		while(fgets(buf, 2048, fp)) {
			guint keyval = -1;
			GdkModifierType modmask = 0;
			
			line++;
			
			if(*buf == '#' || !(p = strstr(buf, "=")))
				continue;
			
			while(buf[strlen(buf)-1] == '\n' || buf[strlen(buf)-1] == '\r')
				buf[strlen(buf)-1] = 0;
			
			if(!parse_keybinding(p+1, &keyval, &modmask)) {
				g_warning("Keybinding is invalid:%d: %s", line, buf);
				continue;
			}
			
			g_hash_table_insert(keys_hash, KEYBIND_HASH_KEY(keyval, modmask),
					g_strndup(buf, p-buf));
			
			*buf = 0;
		}
		fclose(fp);
		ret = TRUE;
	}
	
	return ret;
}

void
xfmedia_keybindings_init()
{
	gchar *file;
	
	g_return_if_fail(!keys_hash);
	
	keys_hash = g_hash_table_new_full(g_direct_hash, g_direct_equal,
			NULL, (GDestroyNotify)g_free);
	
	file = xfce_resource_lookup(XFCE_RESOURCE_CONFIG, KEYBIND_FILE);
	if(!file) {
		g_warning("No valid keybindings file found.");
		return;
	}
	
	if(!xfmedia_keybindings_parse_file(file))
		g_warning("Failed to parse keybindings file (%s)", file);
	
	g_free(file);
}

static gint
keybinding_compare(gconstpointer a, gconstpointer b)
{
	const XfMediaKeybinding *a_k = a, *b_k = b;
	return g_ascii_strcasecmp(a_k->keybind_id, b_k->keybind_id);
}

static void
keys_hash_to_list(gpointer key, gpointer value, gpointer user_data)
{
	guint binding_mask = GPOINTER_TO_UINT(key), keyval, modmask;
	gchar *keybind_id = value;
	GList **list = user_data;
	XfMediaKeybinding *kb = g_new0(XfMediaKeybinding, 1);
	
	KEYBIND_UNHASH(binding_mask, keyval, modmask);
	kb->keybind_id = keybind_id;
	kb->keyval = keyval;
	kb->modmask = modmask;
	
	*list = g_list_insert_sorted(*list, kb, keybinding_compare);
}

GList *
xfmedia_keybindings_get_list()
{
	GList *list = NULL;
	
	g_hash_table_foreach(keys_hash, keys_hash_to_list, &list);
	
	return list;
}

void
xfmedia_keybindings_free_list(GList *list)
{
	GList *l;
	
	for(l = list; l; l = l->next)
		g_free(l->data);
	if(list)
		g_list_free(list);
}

G_CONST_RETURN gchar *
xfmedia_keybindings_lookup(guint keyval, GdkModifierType modmask)
{
	g_return_val_if_fail(keys_hash, NULL);
	
	/* stupid gtk_accelerator_parse()... */
	if(modmask & GDK_SHIFT_MASK)
		keyval = gdk_keyval_to_lower(keyval);
	
	return g_hash_table_lookup(keys_hash, KEYBIND_HASH_KEY(keyval, modmask));
}

void
xfmedia_keybindings_cleanup()
{
	g_return_if_fail(keys_hash);
	
	g_hash_table_destroy(keys_hash);
	keys_hash = NULL;
}
