/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/Xlib.h>

#include <gdk/gdkx.h>
#include <gtk/gtk.h>

#include <libxfce4util/debug.h>
#include <libxfcegui4/icons.h>

#include "xfmedia-common.h"
#include "xfmedia-icon.h"

XfceIconTheme *icon_theme = NULL;

static gchar *
convert_filename(const gchar *filename)
{
	GString *gstr;
	gchar *newfilename, *c;
	
	g_string_new_len(NULL, strlen(filename)+20);
	
	for(c = (gchar *)filename; *c; c++) {
		switch(*c) {
			case '_':
				g_string_append(gstr, " ");
				break;
			case '%':
				c++;
				if(*c == '2') {
					c++;
					if(*c == '0')
						g_string_append(gstr, " ");
					else
						g_string_append_printf(gstr, "%%2%c", *c);
				} else
					g_string_append_printf(gstr, "%%%c", *c);
				break;
			default:
				g_string_append_c(gstr, *c);
				break;
		}
	}
	
	newfilename = gstr->str;
	g_string_free(gstr, FALSE);
	return newfilename;
}

GtkWidget *
xfmedia_custom_button_new_from_pixdata(const gchar *text,
		const GdkPixdata *pixdata)
{
	GtkWidget *btn, *img, *hbox, *lbl;
	GdkPixbuf *pix, *tmppix;
	gint iw, ih;
	
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_widget_show(hbox);
	
	gtk_icon_size_lookup(GTK_ICON_SIZE_BUTTON, &iw, &ih);
	pix = gdk_pixbuf_from_pixdata(pixdata, FALSE, NULL);
	ih = gdk_pixbuf_get_height(pix);
	if(iw != ih) {
		tmppix = gdk_pixbuf_scale_simple(pix, iw, iw, GDK_INTERP_BILINEAR);
		g_object_unref(G_OBJECT(pix));
		pix = tmppix;
	}
	img = gtk_image_new_from_pixbuf(pix);
	g_object_unref(G_OBJECT(pix));
	gtk_widget_show(img);
	gtk_box_pack_start(GTK_BOX(hbox), img, FALSE, FALSE, 0);
	
	lbl = gtk_label_new_with_mnemonic(text);
	gtk_widget_show(lbl);
	gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
	
	btn = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(btn), hbox);
	gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), btn);
	
	return btn;
}

GtkWidget *
xfmedia_custom_button_new(const gchar *text, const gchar *icon)
{
	GtkWidget *btn, *hbox, *img, *lbl;
	GdkPixbuf *pix;
	gint iw, ih;
	
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_widget_show(hbox);
	
	img = gtk_image_new_from_stock(icon, GTK_ICON_SIZE_BUTTON);
	if(!img || gtk_image_get_storage_type(GTK_IMAGE(img)) == GTK_IMAGE_EMPTY) {
		gtk_icon_size_lookup(GTK_ICON_SIZE_BUTTON, &iw, &ih);
		pix = xfce_icon_theme_load(icon_theme, icon, iw);
		if(pix) {
			img = gtk_image_new_from_pixbuf(pix);
			g_object_unref(G_OBJECT(pix));
		}
	}
	if(img) {
		gtk_widget_show(img);
		gtk_box_pack_start(GTK_BOX(hbox), img, FALSE, FALSE, 0);
	}
	
	lbl = gtk_label_new_with_mnemonic(text);
	gtk_widget_show(lbl);
	gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
	
	btn = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(btn), hbox);
	gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), btn);
	
	return btn;
}

gchar *
xfmedia_filename_to_name(const gchar *filename)
{
	gchar *p, *q, *utf8, *tmp;
	const gchar *charset = NULL;
	gint written = 0;
	
	/* FIXME: i don't like this, but i don't want e.g. "dvd://" to return
	 * a blank name */
	if((tmp=strstr(filename, ":/")) && tmp-filename < 6 && !strstr(filename, "://"))
		p = g_strdup(filename);
	else {
		p = g_strrstr(filename, G_DIR_SEPARATOR_S);
		if(p)
			p = g_strdup(p+1);
		else
			p = g_strdup(filename);
		q = strstr(p, "_");
		while(q) {
			*q = ' ';
			q = strstr(q+1, "_");
		}
		q = g_strrstr(p, ".");
		if(q)
			*q = 0;
	}
	
	if(!g_utf8_validate(p, -1, NULL)) {
		if(!g_get_charset(&charset)) {
			utf8 = g_convert_with_fallback(p, strlen(p), "UTF-8", charset, "?",
					NULL, &written, NULL);
		} else
			utf8 = g_locale_to_utf8(p, -1, NULL, &written, NULL);
		if(!utf8)
			return p;
		g_free(p);
		return utf8;
	}
	
	return p;
}

gchar *
xfmedia_sinfo_to_string(xine_stream_t *stream, const gchar *fmt, const gchar *mrl)
{
	GString *gstr;
	gchar *str = NULL, *c, *filename;
	const gchar *value;
	gboolean have_a_field = FALSE;
	
	/*DBG("xfmedia_sinfo_to_string(%p, %s, %s)\n", stream, fmt, mrl);*/
	
	gstr = g_string_new_len(NULL, 128);
	
	for(c = (gchar *)fmt; *c; c++) {
		if(*c == '%') {
			c++;
			switch(*c) {
				case 'p':
					value = xine_get_meta_info(stream, XINE_META_INFO_ARTIST);
					if(value) {
						g_string_append(gstr, value);
						have_a_field = TRUE;
					}
					break;
				case 't':
					value = xine_get_meta_info(stream, XINE_META_INFO_TITLE);
					if(value) {
						g_string_append(gstr, value);
						have_a_field = TRUE;
					}
					break;
				case 'a':
					value = xine_get_meta_info(stream, XINE_META_INFO_ALBUM);
					if(value) {
						g_string_append(gstr, value);
						have_a_field = TRUE;
					}
					break;
				case 'T':
					/* not supported by xine =( */
					break;
				case 'g':
					value = xine_get_meta_info(stream, XINE_META_INFO_GENRE);
					if(value) {
						g_string_append(gstr, value);
						have_a_field = TRUE;
					}
					break;
				case 'y':
					value = xine_get_meta_info(stream, XINE_META_INFO_YEAR);
					if(value) {
						g_string_append(gstr, value);
						have_a_field = TRUE;
					}
					break;
				case 'f':
					if(mrl) {
						filename = convert_filename(mrl);
						g_string_append(gstr, filename);
						g_free(filename);
						have_a_field = TRUE;
					}
					break;
				case '%':
					g_string_append(gstr, "%");
					break;
				default:
					g_string_append_printf(gstr, "%%%c", *c);
					break;
			}
		} else
			g_string_append_c(gstr, *c);
	}
	
	if(have_a_field) {
		str = gstr->str;
		g_string_free(gstr, FALSE);
	} else
		g_string_free(gstr, TRUE);
	
	return str;
}

void
xfmedia_init_icon_theme()
{
	if(icon_theme)
		return;
	
	icon_theme = xfce_icon_theme_get_for_screen(NULL);
	xfce_icon_theme_prepend_search_path(icon_theme, DATADIR "/icons/");
}

gchar *
xfmedia_num_prettify(const gchar *buf)
{
	gchar *p, *np, *str;
	gint newlen, commas, i;
	
	if(!buf)
		return NULL;
	if(strlen(buf) < 4)
		return g_strdup(buf);
	
	commas = (strlen(buf)-1) / 3;
	newlen = strlen(buf) + commas;
	str = g_malloc0(newlen+1);
	
	i = 3;
	for(p=(gchar *)buf+strlen(buf)-1, np=str+newlen-1; p >= buf; p--, np--) {
		*np = *p;
		if(--i == 0) {
			*--np = ',';
			i = 3;
		}
	}
	
	return str;
}

GtkWidget *
xfmedia_gtkimage_load(const gchar *icon_name, GtkIconSize icon_size)
{
	GtkWidget *img;
	GdkPixbuf *tmp, *pix;
	gint iw, ih;
	
	gtk_icon_size_lookup(icon_size, &iw, &ih);
	
	tmp = xfce_icon_theme_load(icon_theme, icon_name, iw);
	if(tmp) {
		if(gdk_pixbuf_get_width(tmp) != iw || gdk_pixbuf_get_height(tmp) != ih) {
			pix = gdk_pixbuf_scale_simple(tmp, iw, ih, GDK_INTERP_BILINEAR);
			g_object_unref(G_OBJECT(tmp));
		} else
			pix = tmp;
		
		img = gtk_image_new_from_pixbuf(pix);
		g_object_unref(G_OBJECT(pix));
	} else
		img = gtk_image_new();
	
	gtk_widget_show(img);

	return img;
}

GdkWindow *
xfmedia_get_window_decorations(GdkWindow *gwindow)
{
	GdkWindow *wmdec_win = NULL;
	GdkDisplay *gdpy = gdk_drawable_get_display(GDK_DRAWABLE(gwindow));
	Display *dpy = GDK_DISPLAY_XDISPLAY(gdpy);
	Window window = GDK_WINDOW_XWINDOW(gwindow);
	Window root_return, parent_return = None, *children_return = NULL;
	unsigned int nchildren = 0;
	
	DBG("looking for decorations for 0x%08x", (guint)window);
	
	for(;;) {
		DBG("checking window 0x%08x", (guint)window);
		XQueryTree(dpy, window, &root_return, &parent_return,
				&children_return, &nchildren);
		XFree(children_return);
		DBG("got parent 0x%08x, =? root 0x%08x", (guint)parent_return, (guint)root_return);
		if(root_return == parent_return)
			break;
		window = parent_return;
	};
	
	if(!(wmdec_win = gdk_window_lookup_for_display(gdpy, window)))
		wmdec_win = gdk_window_foreign_new_for_display(gdpy, window);
	
	DBG("got window decorations: %p (xid: 0x%08x)", wmdec_win, (guint)window);
	
	return wmdec_win;
}

/* ripped off of libgnomesomething, i think. */
#if 1
GList *
xfmedia_uri_list_extract_uris(const gchar *uri_list)
{
	const gchar *p, *q;
	gchar *retval;
	GList *result = NULL;
	
	g_return_val_if_fail(uri_list != NULL, NULL);
	
	p = uri_list;
	
	/* We don't actually try to validate the URI according to RFC
	 * 2396, or even check for allowed characters - we just ignore
	 * comments and trim whitespace off the ends.  We also
	 * allow LF delimination as well as the specified CRLF.
	 */
	while(p) {
		if(*p != '#') {
			while(g_ascii_isspace((int)*p))
				p++;
			
			q = p;
			while(*q && (*q != '\n') && (*q != '\r'))
				q++;
			
			if(q > p) {
				q--;
				while(q > p && g_ascii_isspace((int)*q))
					q--;
				
				retval = (gchar *)g_malloc(q - p + 2);
				strncpy(retval, p, q - p + 1);
				retval[q - p + 1] = '\0';
				
				result = g_list_prepend (result, retval);
			}
		}
		p = strchr (p, '\n');
		if(p)
			p++;
	}
	
	return g_list_reverse(result);
}

#else
GList *
xfmedia_uri_list_extract_uris(const gchar *uri_list)
{
	const gchar *p, *q;
	gchar *retval;
	GList *result = NULL;
	
	g_return_val_if_fail(uri_list != NULL, NULL);
	
	p = uri_list;

	/* We don't actually try to validate the URI according to RFC
	 * 2396, or even check for allowed characters - we just ignore
	 * comments and trim whitespace off the ends.  We also
	 * allow LF delimination as well as the specified CRLF.
	 */
	while(p) {
		if(*p != '#') {
			while(g_ascii_isspace((int)(*p)))
				p++;

			q = p;
			while(*q && (*q != '\n') && (*q != '\r'))
				q++;
			
			if(q > p) {
				q--;
				while(q > p && g_ascii_isspace(((int)(*q))))
					q--;
				
				retval = (gchar *)g_malloc(q - p + 2);
				g_strlcpy(retval, p, q - p + 1);
				retval[q - p + 1] = '\0';
				
				result = g_list_prepend(result, retval);
			}
		}
		p = strchr(p, '\n');
		if(p)
			p++;
	}
	
	return g_list_reverse(result);
}
#endif

	
#if !GLIB_CHECK_VERSION(2, 4, 0)
/* nicked from glib 2.4.5 */
GList *
g_queue_peek_nth_link (GQueue *queue,
                       guint   n)
{
  GList *link;
  gint i;

  g_return_val_if_fail (queue != NULL, NULL);

  if (n >= queue->length)
    return NULL;

  if (n > queue->length / 2)
    {
      n = queue->length - n - 1;

      link = queue->tail;
      for (i = 0; i < n; ++i)
        link = link->prev;
    }
  else
    {
      link = queue->head;
      for (i = 0; i < n; ++i)
        link = link->next;
    }

  return link;
}

void
g_queue_unlink (GQueue *queue,
                GList  *link_)
{
  g_return_if_fail (queue != NULL);
  g_return_if_fail (link_ != NULL);

  if (link_ == queue->tail)
    queue->tail = queue->tail->prev;
  
  queue->head = g_list_remove_link (queue->head, link_);
  queue->length--;
}

gpointer
g_queue_pop_nth (GQueue *queue,
                 guint   n)
{
  GList *nth_link;
  gpointer result;
  
  g_return_val_if_fail (queue != NULL, NULL);

  if (n >= queue->length)
    return NULL;
  
  nth_link = g_queue_peek_nth_link (queue, n);
  result = nth_link->data;
  
  g_queue_unlink (queue, nth_link);
  g_list_free (nth_link);

  return result;
}
#endif

#if !GTK_CHECK_VERSION(2, 4, 0)

#define _NET_WM_STATE_REMOVE 0
#define _NET_WM_STATE_ADD    1

static void
_xfmedia_change_netwm_state(GdkWindow *window, const gchar *state, gboolean add)
{
	XClientMessageEvent xevt;
	Display *dpy = GDK_WINDOW_XDISPLAY(window);
	GdkScreen *gscreen = gdk_drawable_get_screen(GDK_DRAWABLE(window));
	Window root = GDK_WINDOW_XWINDOW(gdk_screen_get_root_window(gscreen));
	
	xevt.type = ClientMessage;
	xevt.serial = 0;
	xevt.send_event = True;
	xevt.window = GDK_WINDOW_XWINDOW(window);
	xevt.message_type = XInternAtom(dpy, "_NET_WM_STATE", False);
	xevt.format = 32;
	xevt.data.l[0] = add ? _NET_WM_STATE_ADD : _NET_WM_STATE_REMOVE;
	xevt.data.l[1] = XInternAtom(dpy, state, False);
	xevt.data.l[2] = 0;
	
	gdk_error_trap_push();
	XSendEvent(dpy, root, False,
			SubstructureRedirectMask|SubstructureNotifyMask, (XEvent *)&xevt);
	gdk_error_trap_pop();
}


void
gtk_window_set_keep_above(GtkWindow *window, gboolean setting)
{
	GdkWindow *the_win = NULL;
	
	if(window->frame)
		the_win = window->frame;
	else if(GTK_WIDGET(window)->window)
		the_win = GTK_WIDGET(window)->window;
	else
		return;
	
	_xfmedia_change_netwm_state(the_win, "_NET_WM_STATE_ABOVE", setting);
}
#endif
